"""ORM client for a specific cell in an inventory in SciNote."""

import logging

from ..client.api.inventory_cell_client import InventoryCellClient
from ..client.models.inventory_cell import (
    CreateInventoryCell,
    UpdateInventoryCell,
    Attributes,
)

logger = logging.getLogger(__name__)


class Cell:
    """ORM client for a specific cell in an inventory in SciNote."""

    def __init__(
        self, cell_id: int, attributes: Attributes, client: InventoryCellClient
    ):
        self.cell_id = cell_id
        self.attributes = attributes
        self.__client = client

    def value(self):
        """Get the value of the cell."""
        logger.debug(f'Getting value for cell {self.attributes}')

        # Empty cells will not have a value.
        if self.attributes.value is None:
            return None

        match self.attributes.value_type:
            case 'text':
                return self.attributes.value.text
            case 'number':
                return self.attributes.value.data
            case 'list':
                return self.attributes.value.inventory_list_item_name
            case 'date_time':
                return self.attributes.value.date_time
            case 'date':
                return self.attributes.value.date
            case _:
                raise ValueError(
                    f'Unknown value type {self.attributes.value_type}'
                )

    async def update_value(self, value: str):
        """Update the value stored in cell."""
        logger.debug(f'Updating value for cell {self.attributes}')
        # We need to check if we are creating a new value or updating an
        # existing one.

        if self.attributes.value is None:
            await self.__do_create_cell(value)
        else:
            await self.__do_update_cell(value)

    async def __do_create_cell(self, value: str):
        try:
            result = await self.__client.create_cell(
                CreateInventoryCell(
                    value=value, column_id=self.attributes.column_id
                )
            )
            if result:
                self.cell_id = result.id
                self.attributes = result.attributes
        except Exception as e:
            logger.error(f'Failed to create cell value={value}: {e}')
            raise

    async def __do_update_cell(self, value: str):
        try:
            result = await self.__client.update_cell(
                UpdateInventoryCell(
                    id=self.cell_id,
                    value=value,
                    column_id=self.attributes.column_id,
                )
            )
            if result:
                self.attributes = result.attributes
        except Exception as e:
            logger.error(f'Failed to update cell {self.cell_id}: {e}')
            raise

    def match(self, value: str) -> bool:
        """Check if the value of the cell matches the supplied string."""

        if self.attributes.value is None:
            return False

        match self.attributes.value_type:
            case 'text':
                return value == self.attributes.value.text
            case 'number':
                return value == self.attributes.value.data
            case 'list':
                return value == self.attributes.value.inventory_list_item_name
            case 'date_time':
                return value == self.attributes.value.date_time
            case 'date':
                return value == self.attributes.value.date
            case _:
                raise ValueError(
                    f'Unknown value type {self.attributes.value_type}'
                )

        return False



"""Model representing a team in SciNote."""

from pydantic import BaseModel, ConfigDict


class InventoryChild(BaseModel):
    model_config = ConfigDict(extra='ignore')

    id: str
    type: str
    relationships: dict



"""SciNote REST API client for inventory column list items."""

import httpx
import logging

from typing import List
from pydantic import ValidationError

from ..models.inventory_column_list_item import (
    InventoryColumnListItem,
    CreateInventoryColumnListItem,
)

logger = logging.getLogger(__name__)


class InventoryColumnListItemClient:
    """SciNote REST API client for inventory column list items."""

    def __init__(
        self,
        base_url: str,
        api_key: str,
        team_id: str,
        inventory_id: str,
        column_id: str,
    ):
        """Initialize the client with the base URL and token."""
        self.base_url = base_url
        self.api_key = api_key
        self.team_id = team_id
        self.inventory_id = inventory_id
        self.column_id = column_id
        self.headers = {
            'Api-Key': api_key,
        }

    async def get_list_items(self) -> List[InventoryColumnListItem]:
        """Get a list of inventory column list items."""
        url = (
            f'{self.base_url}/api/v1/teams/{self.team_id}/inventories/'
            f'{self.inventory_id}/columns/{self.column_id}/list_items'
        )

        result = []

        # Keep track of visited urls to avoid infinite loops
        visited = set()
        async with httpx.AsyncClient() as client:
            while url:
                response = await client.get(url, headers=self.headers)
                response.raise_for_status()
                visited.add(url)
                logger.debug(f'Get columns response: {response.json()}')
                try:
                    [
                        result.append(InventoryColumnListItem(**column))
                        for column in response.json()['data']
                    ]
                except ValidationError as e:
                    raise ValueError(f'Invalid response from SciNote: {e}')

                url = response.json()['links'].get('next', None)

                # Check we haven't visited this url before
                if url in visited:
                    break

            return result

    async def create_list_item(
        self, item: CreateInventoryColumnListItem
    ) -> InventoryColumnListItem:
        """Create a new list item."""
        url = (
            f'{self.base_url}/api/v1/teams/{self.team_id}/inventories/'
            f'{self.inventory_id}/columns/{self.column_id}/list_items'
        )

        async with httpx.AsyncClient() as client:
            response = await client.post(
                url,
                json={
                    'data': {
                        'type': item.type,
                        'attributes': {
                            'data': item.data,
                        },
                    },
                },
                headers=self.headers,
            )
            response.raise_for_status()
            try:
                return InventoryColumnListItem(**response.json()['data'])
            except ValidationError as e:
                logger.error(f'Invalid response from SciNote: {e}')
                raise



"""Maps a SciNote item to an ORM item."""

import logging

from datetime import datetime

from .cell import Cell
from ..client.api.inventory_item_client import InventoryItemClient
from ..client.models.inventory_cell import InventoryCell, Attributes

logger = logging.getLogger(__name__)
"""
An item is a complete row in the inventory.

When SciNote returns an item, the values are mapped to a column ID. We need to
do this mapping here using the columns for the inventory and the values that
are passed into the constructor.
"""


class Item:
    """Maps a SciNote item to an ORM item."""

    def __init__(
        self,
        id: str,
        name: str,
        created_at: datetime,
        item_client: InventoryItemClient,
        columns: dict,
        values: list[InventoryCell],
    ):
        self.id = id
        self.name = name
        self.created_at = created_at
        self.item_client = item_client
        self.column_ids = {}
        self.column_names = {}
        self.__cells = {}

        logger.debug(f'Columns: {columns}')
        logger.debug(f'Values: {values}')

        # Create a map of column ids to names and vice versa.
        for index, (key, value) in enumerate(columns.items()):
            name = key.lower().replace(' ', '_')
            self.column_ids[value.id] = name
            self.column_names[name] = value.id

        # For each incoming value, map it to the column name. There is added
        # complexity as the name of the field that is the value of the cell
        # is dependent on the type of the column.
        for value in values:
            column_id = str(value.attributes.column_id)
            column_name = self.column_ids[column_id]
            self.__cells[column_name] = Cell(
                value.id,
                value.attributes,
                item_client.cell_client(self.id),
            )

        # SciNote will only return values if they are not empty. We need to
        # create empty cells for the columns that are not present.
        empty_columns = set(self.column_names.keys()) - set(self.__cells.keys())
        logger.debug(f'Empty columns: {empty_columns}')
        for column_name in empty_columns:
            # Get metadata for the missing column.
            column = columns[column_name]
            self.__cells[column_name] = Cell(
                None,  # No id for cells that do not exist.
                Attributes(
                    column_id=column.id,
                    value_type=column.attributes.data_type,
                ),
                item_client.cell_client(self.id),
            )

    def cell(self, name) -> Cell:
        """Get the item by name."""
        if self.column_names.get(name) is None:
            raise ValueError(f'Item {name} not found')
        return self.__cells.get(name)

    def cells(self) -> list[Cell]:
        """Get all of the cells."""
        return [value for value in self.__cells.values()]

    def match(self, **kwargs) -> bool:
        """Match the item against the given values."""
        for key, value in kwargs.items():
            if key == 'name':
                if self.name != value:
                    return False
                continue
            if key == 'id':
                if self.id != value:
                    return False
                continue

            if key not in self.column_names:
                logger.warning(f'Column {key} not found in item')
                return False

            if not self.cell(key).match(value):
                return False

        return True



"""Maps all of the inventories for a SciNote team."""

import logging

from aiocache import cached

from .inventory import Inventory
from ..client.api.inventory_client import InventoryClient

logger = logging.getLogger(__name__)

# How long we cache result from SciNote before refreshing.
CACHE_TIMEOUT_SECONDS = 120


class Inventories:
    """Maps all of the inventories for a SciNote team."""

    def __init__(self, client: InventoryClient):
        self.__client = client

    async def inventory(self, name: str) -> Inventory:
        """Get the inventory by name."""

        inventories = await self.__load_inventories()

        logger.debug(f'Checking for inventory {name}')
        if inventories.get(name) is None:
            raise ValueError(f'Inventory {name} not found')

        return inventories.get(name)

    async def inventories(self) -> list[Inventory]:
        """Get all of the inventories."""
        inventories = await self.__load_inventories()
        return [value for value in inventories.values()]

    @cached(ttl=CACHE_TIMEOUT_SECONDS)
    async def __load_inventories(self):
        logger.debug('Loading all inventories')

        result = {}
        scinote_inventories = await self.__client.get_inventories()
        for inventory in scinote_inventories:
            inventory_name = inventory.attributes.name.lower().replace(' ', '_')
            logger.debug(f'Adding inventory {inventory_name}')
            result[inventory_name] = Inventory(
                inventory_name,
                self.__client.column_client(inventory.id),
                self.__client.item_client(inventory.id),
            )

        return result



"""SciNote rest api client for inventory cells."""

import httpx

from typing import List
from pydantic import ValidationError

from ..models.inventory_cell import (
    InventoryCell,
    CreateInventoryCell,
    UpdateInventoryCell,
)


class InventoryCellClient:
    """Client for interacting with the SciNote REST API."""

    def __init__(
        self,
        base_url: str,
        api_key: str,
        team_id: int,
        inventory_id: int,
        item_id: int,
    ):
        """Initialize the client."""
        self.base_url = base_url
        self.api_key = api_key
        self.team_id = team_id
        self.inventory_id = inventory_id
        self.item_id = item_id
        self.headers = {
            'Api-Key': api_key,
        }

    async def get_cells(self) -> List[InventoryCell]:
        """Get all inventory cells from SciNote."""

        # Need to loop through all pages to get all of the columns
        url = (
            f'{self.base_url}/api/v1/teams/{self.team_id}/inventories/'
            f'{self.inventory_id}/items/{self.item_id}/cells'
        )
        result = []

        # Keep track of visited urls to avoid infinite loops
        visited = set()
        async with httpx.AsyncClient() as client:
            while url:
                response = await client.get(url, headers=self.headers)
                response.raise_for_status()
                visited.add(url)
                try:
                    [
                        result.append(InventoryCell(**column))
                        for column in response.json()['data']
                    ]
                except ValidationError as e:
                    raise ValueError(f'Invalid response from SciNote: {e}')

                url = response.json()['links'].get('next', None)

                # Check we haven't visited this url before
                if url in visited:
                    break

            return result

    async def create_cell(self, cell: CreateInventoryCell) -> InventoryCell:
        """Create a new inventory cell in SciNote."""
        url = (
            f'{self.base_url}/api/v1/teams/{self.team_id}/inventories/'
            f'{self.inventory_id}/items/{self.item_id}/cells'
        )
        async with httpx.AsyncClient() as client:
            response = await client.post(
                url,
                headers=self.headers,
                json={
                    'data': {
                        'type': cell.type,
                        'attributes': {
                            'value': cell.value,
                            'column_id': cell.column_id,
                        },
                    },
                },
            )
            response.raise_for_status()
            return InventoryCell(**response.json()['data'])

    async def update_cell(self, cell: UpdateInventoryCell) -> InventoryCell:
        """Update an inventory cell in SciNote."""
        url = (
            f'{self.base_url}/api/v1/teams/{self.team_id}/inventories/'
            f'{self.inventory_id}/items/{self.item_id}/cells/{cell.id}'
        )
        async with httpx.AsyncClient() as client:
            response = await client.patch(
                url,
                headers=self.headers,
                json={
                    'data': {
                        'id': cell.id,
                        'type': cell.type,
                        'attributes': {
                            'value': cell.value,
                            'column_id': cell.column_id,
                        },
                    },
                },
            )
            response.raise_for_status()
            # If there was no update then the server will return a 204 with no
            # body, so we will return None in that case.
            if response.status_code == 204:
                return None
            else:
                return InventoryCell(**response.json()['data'])



"""ORM client for an specific inventory in SciNote."""

import logging

from aiocache import Cache

from .items import Item
from ..client.api.inventory_column_client import InventoryColumnClient
from ..client.api.inventory_item_client import InventoryItemClient
from ..client.models.inventory_cell import CreateInventoryCell

logger = logging.getLogger(__name__)

# How long we cache result from SciNote before refreshing.
CACHE_TIMEOUT_SECONDS = 120


class Inventory:
    """ORM client for a specific inventory in SciNote."""

    def __init__(
        self,
        name: str,
        column_client: InventoryColumnClient,
        item_client: InventoryItemClient,
    ):
        self.column_client = column_client
        self.item_client = item_client
        self.name = name
        self.__columns = {}
        self.__item_list = []
        self.__cache = Cache(Cache.MEMORY, ttl=CACHE_TIMEOUT_SECONDS)

    async def load_columns(self) -> None:
        """Load the columns for this inventory."""

        if not await self.__cache.exists('columns'):
            columns = await self.column_client.get_columns()
            self.__columns = {}
            for column in columns:
                name = column.attributes.name.lower().replace(' ', '_')
                self.__columns[name] = column
            await self.__cache.set('columns', True)

    async def load_items(self) -> None:
        """Load the items for this inventory."""

        if not await self.__cache.exists('items'):
            items = await self.item_client.get_items()
            self.__item_list = []
            for item in items:
                self.__item_list.append(
                    Item(
                        item.id,
                        item.attributes.name,
                        item.attributes.created_at,
                        self.item_client,
                        self.__columns,
                        item.inventory_cells,
                    )
                )
            await self.__cache.set('items', True)

    async def items(self) -> list[Item]:
        """Get the items for this inventory."""
        await self.load_columns()
        await self.load_items()
        return self.__item_list

    async def match(self, **kwargs) -> list[Item]:
        """Return matching items from this inventory."""
        await self.load_columns()
        await self.load_items()
        return [item for item in self.__item_list if item.match(**kwargs)]

    async def columns(self):
        """Get the columns for this inventory."""
        await self.load_columns()
        return [column for column in self.__columns.values()]

    async def has_column(self, name: str) -> bool:
        """Check if the inventory has a column."""
        await self.load_columns()
        return name in self.__columns

    async def create_item(self, name: str, **kwargs) -> Item:
        """Create a new item in this inventory."""
        await self.load_columns()

        cells = []
        for key, value in kwargs.items():
            if key not in self.__columns:
                raise ValueError(f'Column {key} does not exist in inventory.')

            column = self.__columns[key]
            cells.append(CreateInventoryCell(value=value, column_id=column.id))

        item = await self.item_client.create_item(name, cells)

        new_item = Item(
            item.id,
            item.attributes.name,
            item.attributes.created_at,
            self.item_client,
            self.__columns,
            item.inventory_cells,
        )
        self.__item_list.append(new_item)
        return new_item
